import JSBI from 'jsbi';

function _typeof(obj) {
  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _classCallCheck(instance, Constructor) {
  if (!(instance instanceof Constructor)) {
    throw new TypeError("Cannot call a class as a function");
  }
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);
    if (enumerableOnly) symbols = symbols.filter(function (sym) {
      return Object.getOwnPropertyDescriptor(object, sym).enumerable;
    });
    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(source, true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(source).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = new Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  }
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArray(iter) {
  if (Symbol.iterator in Object(iter) || Object.prototype.toString.call(iter) === "[object Arguments]") return Array.from(iter);
}

function _iterableToArrayLimit(arr, i) {
  if (!(Symbol.iterator in Object(arr) || Object.prototype.toString.call(arr) === "[object Arguments]")) {
    return;
  }

  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance");
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance");
}

function eq(comparahend, comparator) {
  return comparahend === comparator || is_zero(sub(comparahend, comparator));
}
function lt(comparahend, comparator) {
  return is_negative(sub(comparahend, comparator));
}
function lte(comparahend, comparator) {
  return lt(comparahend, comparator) || eq(comparahend, comparator);
}
function gt(comparahend, comparator) {
  return lt(comparator, comparahend);
}
function gte(comparahend, comparator) {
  return gt(comparahend, comparator) || eq(comparahend, comparator);
}

var relational = /*#__PURE__*/Object.freeze({
  __proto__: null,
  eq: eq,
  lt: lt,
  lte: lte,
  gt: gt,
  gte: gte
});

function is_big_float(big) {
  return _typeof(big) === "object" && big.coefficient instanceof JSBI && Number.isSafeInteger(big.exponent);
}
function is_number(token) {
  return !Number.isNaN(Number(token));
}
function is_negative(big) {
  return JSBI.LT(big.coefficient, BIGINT_ZERO);
}
function is_positive(big) {
  return JSBI.GE(big.coefficient, BIGINT_ZERO);
}
function is_zero(big) {
  return JSBI.EQ(big.coefficient, BIGINT_ZERO);
}
function is_integer(a) {
  return eq(a, integer(a));
}

var predicates = /*#__PURE__*/Object.freeze({
  __proto__: null,
  is_big_float: is_big_float,
  is_number: is_number,
  is_negative: is_negative,
  is_positive: is_positive,
  is_zero: is_zero,
  is_integer: is_integer
});

function make_big_float(coefficient, exponent) {
  if (JSBI.EQ(coefficient, BIGINT_ZERO)) {
    return ZERO;
  }

  var new_big_float = Object.create(null);
  new_big_float.coefficient = coefficient;
  new_big_float.exponent = exponent;
  return Object.freeze(new_big_float);
}
function number(a) {
  if (typeof a !== "number" && typeof a !== "string") {
    if (a instanceof JSBI) {
      return JSBI.toNumber(a);
    } else if (is_big_float(a)) {
      return a.exponent === 0 ? JSBI.toNumber(a.coefficient) : JSBI.toNumber(a.coefficient) * Math.pow(10, a.exponent);
    }
  }

  return Number(a);
}
function normalize(a) {
  var coefficient = a.coefficient,
      exponent = a.exponent; // If the exponent is zero, it is already normal.

  if (exponent !== 0) {
    if (exponent > 0) {
      coefficient = JSBI.multiply(coefficient, JSBI.exponentiate(BIGINT_TEN, JSBI.BigInt(exponent)));
      exponent = 0;
    } else {
      var quotient;
      var remainder; // tslint:disable-next-line: no-bitwise

      while (exponent <= -7) {
        quotient = JSBI.divide(coefficient, BIGINT_TEN_MILLION);
        remainder = JSBI.remainder(coefficient, BIGINT_TEN_MILLION);

        if (!JSBI.EQ(remainder, BIGINT_ZERO)) {
          break;
        }

        coefficient = quotient;
        exponent += 7;
      }

      while (exponent < 0) {
        quotient = JSBI.divide(coefficient, BIGINT_TEN);
        remainder = JSBI.remainder(coefficient, BIGINT_TEN);

        if (!JSBI.EQ(remainder, BIGINT_ZERO)) {
          break;
        }

        coefficient = quotient;
        exponent += 1;
      }
    }
  }

  return make_big_float(coefficient, exponent);
}
function integer(a) {
  // The integer function is like the normalize function except that it throws
  // away significance. It discards the digits after the decimal point.
  var coefficient = a.coefficient,
      exponent = a.exponent; // If the exponent is zero, it is already an integer.

  if (exponent === 0) {
    return a;
  } // If the exponent is positive,
  // multiply the coefficient by 10 ** exponent.


  if (exponent > 0) {
    return make_big_float(JSBI.multiply(coefficient, JSBI.exponentiate(BIGINT_TEN, JSBI.BigInt(exponent))), 0);
  } // If the exponent is negative, divide the coefficient by 10 ** -exponent.
  // This truncates the unnecessary bits. This might be a zero result.


  return make_big_float(JSBI.divide(coefficient, JSBI.exponentiate(BIGINT_TEN, JSBI.BigInt(-exponent))), 0);
}
function fraction(a) {
  return sub(a, integer(a));
}
function make(a, b) {
  var number_pattern = /^(-?\d+)(?:\.(\d*))?(?:e(-?\d+))?$/; // . Capturing groups
  // .      [1] int
  // .      [2] frac
  // .      [3] exp

  if (a instanceof JSBI) {
    return make_big_float(a, Number(b) || 0);
  } else if (typeof a === "string" || typeof a === "number") {
    a = String(a);

    if (Number.isSafeInteger(Number(b))) {
      return make(JSBI.BigInt(parseInt(a, Number(b))), 0);
    }

    var parts = a.match(number_pattern);

    if (parts) {
      var frac = parts[2] || "";
      return make(JSBI.BigInt(parts[1] + frac), (Number(parts[3]) || 0) - frac.length);
    }
  } else if (is_big_float(a)) {
    return a;
  }

  return ZERO;
}
function string(a, radix) {
  if (is_zero(a)) {
    return "0";
  }

  if (radix && is_big_float(radix)) {
    radix = normalize(radix);
    return radix && radix.exponent === 0 ? integer(a).coefficient.toString(JSBI.toNumber(radix.coefficient)) : undefined;
  }

  a = normalize(a);
  var s = (JSBI.LT(a.coefficient, BIGINT_ZERO) ? -a.coefficient : a.coefficient).toString();

  if (a.exponent < 0) {
    var point = s.length + a.exponent;

    if (point <= 0) {
      s = "0".repeat(1 - point) + s;
      point = 1;
    }

    s = s.slice(0, point) + "." + s.slice(point);
  } else if (a.exponent > 0) {
    s += "0".repeat(a.exponent);
  }

  if (JSBI.LT(a.coefficient, BIGINT_ZERO)) {
    s = "-" + s;
  }

  return s;
}
function scientific(a) {
  if (is_zero(a)) {
    return "0";
  }

  a = normalize(a);
  var s = String(JSBI.LT(a.coefficient, BIGINT_ZERO) ? -a.coefficient : a.coefficient);
  var e = a.exponent + s.length - 1;

  if (s.length > 1) {
    s = s.slice(0, 1) + "." + s.slice(1);
  }

  if (e !== 0) {
    s += "e" + e;
  }

  if (JSBI.LT(a.coefficient, BIGINT_ZERO)) {
    s = "-" + s;
  }

  return s;
}

var constructors = /*#__PURE__*/Object.freeze({
  __proto__: null,
  make_big_float: make_big_float,
  number: number,
  normalize: normalize,
  integer: integer,
  fraction: fraction,
  make: make,
  string: string,
  scientific: scientific
});

var BIGINT_ZERO = JSBI.BigInt("0");
var BIGINT_ONE = JSBI.BigInt("1");
var BIGINT_TEN = JSBI.BigInt("10");
var BIGINT_TEN_MILLION = JSBI.BigInt("10000000");
var ZERO = Object.create(null);
ZERO.coefficient = BIGINT_ZERO;
ZERO.exponent = 0;
Object.freeze(ZERO);
var PRECISION = -4;
function set_precision(n) {
  n = Number(n);

  if (!Number.isInteger(n) || Number(n) >= 0) {
    throw Error("Only negative integers are allowed for precision.");
  }

  PRECISION = n;
}
var EPSILON = make("0.0000000000000000000000000000000000000000000000001");
var ONE = make("1");
var TWO = make("2");

function neg(a) {
  return make_big_float(JSBI.unaryMinus(a.coefficient), a.exponent);
}
function abs(a) {
  return is_negative(a) ? neg(a) : a;
}

function conform_op(op) {
  return function (a, b) {
    var differential = a.exponent - b.exponent;
    return differential === 0 ? make_big_float(op(a.coefficient, b.coefficient), a.exponent) : differential > 0 ? make_big_float(op(JSBI.multiply(a.coefficient, JSBI.exponentiate(BIGINT_TEN, JSBI.BigInt(differential))), b.coefficient), b.exponent) : make_big_float(op(a.coefficient, JSBI.multiply(b.coefficient, JSBI.exponentiate(BIGINT_TEN, JSBI.BigInt(-differential)))), a.exponent);
  };
}

var add = conform_op(JSBI.add);
var sub = conform_op(JSBI.subtract);
function mul(multiplicand, multiplier) {
  return make_big_float(JSBI.multiply(multiplicand.coefficient, multiplier.coefficient), multiplicand.exponent + multiplier.exponent);
}
function div(dividend, divisor) {
  var precision = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : PRECISION;

  if (is_zero(dividend) || is_zero(divisor)) {
    return ZERO;
  }

  var coefficient = dividend.coefficient,
      exponent = dividend.exponent;
  exponent -= divisor.exponent;

  if (typeof precision !== "number") {
    precision = number(precision);
  }

  if (exponent > precision) {
    coefficient = JSBI.multiply(coefficient, JSBI.exponentiate(BIGINT_TEN, JSBI.BigInt(exponent - precision)));
    exponent = precision;
  }

  coefficient = JSBI.divide(coefficient, divisor.coefficient);
  return make_big_float(coefficient, exponent);
}
function sqrt(n) {
  var x = n;
  var y = ONE;

  while (gt(sub(x, y), EPSILON)) {
    x = div(add(x, y), TWO);
    y = div(n, x);
  }

  return x;
}
function exponentiation(base, exp) {
  if (eq(exp, ZERO)) {
    return ONE;
  }

  if (is_negative(exp)) {
    return div(ONE, exponentiation(base, abs(exp)));
  }

  if (exp.exponent === 0) {
    var result = base;
    var n = 1;

    while (n !== number(exp)) {
      result = mul(result, base);
      n += 1;
    }

    return result;
  }

  if (gt(exp, ONE) || eq(exp, ONE)) {
    var temp = exponentiation(base, div(exp, TWO));
    return mul(temp, temp);
  }

  var low = ZERO;
  var high = ONE;
  var sqr = sqrt(base);
  var acc = sqr;
  var mid = div(high, TWO);

  while (gt(abs(sub(mid, exp)), EPSILON)) {
    sqr = sqrt(sqr);

    if (lt(mid, exp) || eq(mid, exp)) {
      low = mid;
      acc = mul(acc, sqr);
    } else {
      high = mid;
      acc = mul(acc, div(ONE, sqr));
    }

    mid = div(add(low, high), TWO);
  }

  return acc;
}
function ceil(n) {
  if (is_integer(n)) {
    return n;
  } else {
    return make_big_float(JSBI.add(integer(n).coefficient, BIGINT_ONE), 0);
  }
}
function floor(n) {
  return integer(n);
}

var arithmetic = /*#__PURE__*/Object.freeze({
  __proto__: null,
  neg: neg,
  abs: abs,
  add: add,
  sub: sub,
  mul: mul,
  div: div,
  sqrt: sqrt,
  exponentiation: exponentiation,
  ceil: ceil,
  floor: floor
});

function evaluate(source) {
  var precision = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : PRECISION;

  if (typeof source !== "string") {
    throw Error("The first parameter was expected to be a string.");
  } // This function relies on an algorithm that fully parenthesizes the expression


  function parenthesize(expr) {
    return "((((" + expr.replace(/\(/g, "((((").replace(/\)/g, "))))").replace(/(^|[^!])===?/g, ")))==(((").replace(/<=/g, ")))<=(((").replace(/>=/g, ")))>=(((").replace(/<(?!=)/g, ")))<(((").replace(/>(?!=)/g, ")))>(((").replace(/!==?/g, ")))!=(((").replace(/(^|[^e])\+/g, "))+((").replace(/(^|[^e])-(?!\d)/g, "))-((").replace(/\^|\*\*/g, "**").replace(/(^|[^*])\*(?!\*)/g, ")*(").replace(/\//g, ")/(").replace(/%/g, ")%(").replace(/ /g, "") + "))))";
  }

  var expression = parenthesize(source);
  var rx_tokens = /(-?\d+(?:\.\d+)?(?:e(-?|\+?)\d+)?)|(\(|\))|(\+|-|\/|\*\*|==|!=|<=?|>=?|\*|\^|%)/g; // Capture groups
  // [1] Number
  // [2] Paren
  // [3] Operator
  // Tokenize the expression

  var tokens = (expression.match(rx_tokens) || []).map(function (element) {
    var parens = ["(", ")"];
    var operators = ["+", "-", "*", "**", "/", "%", "==", "!=", "<", ">", "<=", ">="];

    if (element === "%") {
      throw Error("The modulo operator is not supported yet.");
    }

    if (parens.includes(element)) {
      return {
        type: "paren",
        value: element
      };
    } else if (operators.includes(element)) {
      return {
        type: "operator",
        value: element
      };
    } else if (is_number(element)) {
      return {
        type: "number",
        value: normalize(make(element.replace("+", "")))
      };
    }

    var error = 'Unexpected token "' + element + '"';
    throw Error(error);
  });
  var n = 0;
  tokens.forEach(function (element, index) {
    if (element.value === "**") {
      if (tokens[index + 2].value === "**") {
        tokens.splice(index + 1, 0, {
          value: "(",
          type: "paren"
        });
        n += 1;
      } else {
        while (n) {
          n -= 1;
          tokens.splice(index + 3, 0, {
            value: ")",
            type: "paren"
          });
        }
      }
    }
  }); // Recursively resolve the parentheses

  function resolve(arr) {
    // Remove parens when there's only one value
    if (arr.length <= 3) {
      return [arr[1]];
    }

    var last_left_paren;
    var i = 0;

    while (i <= arr.length) {
      var value = arr[i].value;

      if (value === "(") {
        last_left_paren = i;
      }

      if (value === ")") {
        var start = arr.slice(0, last_left_paren);
        var term = arr.splice(last_left_paren, i - last_left_paren + 1);
        var end = arr.slice(last_left_paren, arr.length);
        return resolve([].concat(_toConsumableArray(start), _toConsumableArray(resolve(term)), _toConsumableArray(end)));
      }

      if (arr[i].type === "operator" && arr[i + 1].type !== "paren") {
        var _ret = function () {
          var start = arr.slice(0, arr[i + 2].type === "operator" || arr[i + 2].type === "paren" ? last_left_paren + 1 : last_left_paren);
          var ops = arr.splice(i - 1, 3);
          var end = arr.slice(arr[i - 1].type === "operator" || (arr[i + 1] || {}).type === "paren" || i >= arr.length ? i - 1 : i, arr.length);
          var a = ops[0].value;
          var b = ops[2].value;
          var operator = ops[1].value;
          var bigfloat_return = {
            "+": function _() {
              return add(a, b);
            },
            "-": function _() {
              return sub(a, b);
            },
            "*": function _() {
              return mul(a, b);
            },
            "/": function _() {
              return is_zero(b) ? ZERO : div(a, b, precision);
            },
            "**": function _() {
              return exponentiation(a, b);
            }
          }[operator];
          var boolean_return = {
            "==": function _() {
              return eq(a, b);
            },
            "!=": function _() {
              return !eq(a, b);
            },
            "<": function _() {
              return lt(a, b);
            },
            ">": function _() {
              return gt(a, b);
            },
            "<=": function _() {
              return lt(a, b) || eq(a, b);
            },
            ">=": function _() {
              return gt(a, b) || eq(a, b);
            }
          }[operator];
          var res = bigfloat_return ? {
            type: "number",
            value: bigfloat_return()
          } : {
            type: "boolean",
            value: boolean_return()
          };
          return {
            v: resolve([].concat(_toConsumableArray(start), [res], _toConsumableArray(end)))
          };
        }();

        if (_typeof(_ret) === "object") return _ret.v;
      }

      i += 1;
    }
  }

  var _resolve = resolve(tokens),
      _resolve2 = _slicedToArray(_resolve, 1),
      result = _resolve2[0];

  if (result.type === "number") {
    return string(result.value);
  }

  return result.value;
}

var Decimal = function Decimal(n) {
  var _this = this;

  _classCallCheck(this, Decimal);

  this.setPrecision = set_precision;

  this.decimalPlaces = function () {
    return -_this.exponent;
  };

  this.dp = function () {
    return -_this.exponent;
  };

  this.isInteger = function () {
    return is_integer(_this);
  };

  this.isInt = function () {
    return is_integer(_this);
  };

  this.isNegative = function () {
    return is_negative(_this);
  };

  this.isNeg = function () {
    return is_negative(_this);
  };

  this.isPositive = function () {
    return is_positive(_this);
  };

  this.isPos = function () {
    return is_positive(_this);
  };

  this.isZero = function () {
    return is_zero(_this);
  };

  this.toString = function () {
    return string(_this);
  };

  this.equals = function (y) {
    return eq(_this, make(y));
  };

  this.eq = function (y) {
    return eq(_this, make(y));
  };

  this.greaterThan = function (y) {
    return gt(_this, make(y));
  };

  this.gt = function (y) {
    return gt(_this, make(y));
  };

  this.greaterThanOrEqualTo = function (y) {
    return gte(_this, make(y));
  };

  this.gte = function (y) {
    return gte(_this, make(y));
  };

  this.lessThan = function (y) {
    return lt(_this, make(y));
  };

  this.lt = function (y) {
    return lt(_this, make(y));
  };

  this.lessThanOrEqualTo = function (y) {
    return lte(_this, make(y));
  };

  this.lte = function (y) {
    return lte(_this, make(y));
  };

  this.absoluteValue = function () {
    return new Decimal(abs(_this));
  };

  this.abs = function () {
    return new Decimal(abs(_this));
  };

  this.negated = function () {
    return new Decimal(neg(_this));
  };

  this.neg = function () {
    return new Decimal(neg(_this));
  };

  this.squareRoot = function () {
    return new Decimal(sqrt(_this));
  };

  this.sqrt = function () {
    return new Decimal(sqrt(_this));
  };

  this.dividedBy = function (y) {
    return new Decimal(div(_this, make(y)));
  };

  this.div = function (y) {
    return new Decimal(div(_this, make(y)));
  };

  this.minus = function (y) {
    return new Decimal(sub(_this, make(y)));
  };

  this.sub = function (y) {
    return new Decimal(sub(_this, make(y)));
  };

  this.plus = function (y) {
    return new Decimal(add(_this, make(y)));
  };

  this.add = function (y) {
    return new Decimal(add(_this, make(y)));
  };

  this.times = function (y) {
    return new Decimal(mul(_this, make(y)));
  };

  this.mul = function (y) {
    return new Decimal(mul(_this, make(y)));
  };

  this.toPower = function (y) {
    return new Decimal(exponentiation(_this, make(y)));
  };

  this.pow = function (y) {
    return new Decimal(exponentiation(_this, make(y)));
  };

  this.ceil = function () {
    return new Decimal(ceil(_this));
  };

  this.floor = function () {
    return new Decimal(floor(_this));
  };

  var _make = make(n),
      exponent = _make.exponent,
      coefficient = _make.coefficient;

  this.exponent = exponent;
  this.coefficient = coefficient;
};

var mod = Object.freeze(_objectSpread2({
  BigFloat: make,
  evaluate: evaluate,
  set_precision: set_precision
}, arithmetic, {}, predicates, {}, constructors, {}, relational));

export default mod;
export { Decimal };
